function [THETA_HAT,GRADIENT] = estimator_family(APOSTPDF_SAMPLES, LOG_APOSTPDF_VALS, FPARAMS, LEARNING_PARAMS)
% ESTIMATOR_FAMILY - returns the result of the estimator family for the
% observation using the a posteriori samples APOSTPDF_SAMPLES and the
% parameters FPARAMS.XI1, FPARAMS.XI2, FPARAMS.XI3, FPARAMS.LAMBDA, FPARAMS.PHI, FPARAMS.PSI
%
% Note that X can be a matrix, each column is then one observation
%
% IMPORTANT: estimator_family_core assumes that LOG_APOSTPDF_VALS is already normalized
% otherwise, it may happen that NaNs occur
%
% written by Stefan Uhlich

[M,K] = size(APOSTPDF_SAMPLES{1}); % M = number of unkown parameters, K = number of a posteriori samples
NUM_OBSERVATIONS = numel(APOSTPDF_SAMPLES);



%% CALC ESTIMATES AND (IF NEEDED) GRADIENT
if length(FPARAMS.PHI) == 1 % the mex-File currently only supports scalar phi and psi values
    if nargout == 1
        THETA_HAT = estimator_family_core(APOSTPDF_SAMPLES,LOG_APOSTPDF_VALS, ...
            FPARAMS.XI1, FPARAMS.XI2, FPARAMS.XI3, FPARAMS.LAMBDA, FPARAMS.PHI, FPARAMS.PSI);
    else
        [THETA_HAT,GRADIENT] = estimator_family_core(APOSTPDF_SAMPLES,LOG_APOSTPDF_VALS, ...
            FPARAMS.XI1, FPARAMS.XI2, FPARAMS.XI3, FPARAMS.LAMBDA, FPARAMS.PHI, FPARAMS.PSI);
    end
else
    if nargout == 1
        THETA_HAT = internal_estimator_family_core(APOSTPDF_SAMPLES,LOG_APOSTPDF_VALS, FPARAMS, M);
    else
        [THETA_HAT,GRADIENT] = internal_estimator_family_core(APOSTPDF_SAMPLES,LOG_APOSTPDF_VALS,FPARAMS, M);
    end
end

if LEARNING_PARAMS.DEBUG == 1
    % 1. Compare MEX and INTERNAL
    [THETA_HAT_,GRADIENT_] = internal_estimator_family_core(APOSTPDF_SAMPLES,LOG_APOSTPDF_VALS, FPARAMS, M);
%     [THETA_HAT_,GRADIENT_] = estimator_family_core_old(APOSTPDF_SAMPLES,LOG_APOSTPDF_VALS, ...
%             FPARAMS.XI1, FPARAMS.XI2, FPARAMS.XI3, FPARAMS.LAMBDA, FPARAMS.PHI, FPARAMS.PSI);
    fprintf('DEBUG: Comparison of estimates: Max difference (relative): %e\n',max(max(abs(THETA_HAT-THETA_HAT_))./max(max(abs(THETA_HAT_)))))

    if nargout == 2  && LEARNING_PARAMS.CHECK_EF_GRADIENT == 0 % compare only internal version with mex version
        fprintf('DEBUG: Comp. of gradients (XI1): Max difference (rel): %0.2e / (abs): %0.2e\n',max(max(abs(GRADIENT.XI1-GRADIENT_.XI1)./abs(GRADIENT_.XI1))),max(max(abs(GRADIENT.XI1-GRADIENT_.XI1))))
        fprintf('DEBUG: Comp. of gradients (XI2): Max difference (rel): %0.2e / (abs): %0.2e\n',max(max(abs(GRADIENT.XI2-GRADIENT_.XI2)./abs(GRADIENT_.XI2))),max(max(abs(GRADIENT.XI2-GRADIENT_.XI2))))
        fprintf('DEBUG: Comp. of gradients (XI3): Max difference (rel): %0.2e / (abs): %0.2e\n',max(max(abs(GRADIENT.XI3-GRADIENT_.XI3)./abs(GRADIENT_.XI3))),max(max(abs(GRADIENT.XI3-GRADIENT_.XI3))))
        fprintf('DEBUG: Comp. of gradients (LAMBDA): Max difference (rel): %0.2e / (abs): %0.2e\n',max(max(abs(GRADIENT.LAMBDA-GRADIENT_.LAMBDA)./abs(GRADIENT_.LAMBDA))),max(max(abs(GRADIENT.LAMBDA-GRADIENT_.LAMBDA))))
        fprintf('DEBUG: Comp. of gradients (PHI): Max difference (rel): %0.2e / (abs): %0.2e\n',max(max(abs(GRADIENT.PHI-GRADIENT_.PHI)./abs(GRADIENT_.PHI))),max(max(abs(GRADIENT.PHI-GRADIENT_.PHI))))
        fprintf('DEBUG: Comp. of gradients (PSI): Max difference (rel): %0.2e / (abs): %0.2e\n',max(max(abs(GRADIENT.PSI-GRADIENT_.PSI)./abs(GRADIENT_.PSI))),max(max(abs(GRADIENT.PSI-GRADIENT_.PSI))))
    end

    if nargout == 2 && LEARNING_PARAMS.CHECK_EF_GRADIENT == 1 % also use finite difference approximation for comparison
        addpath('/home/m109/work/Matlab/Tools/DERIVESTsuite')
        TMP_LEARNING_PARAMS = LEARNING_PARAMS;
        TMP_LEARNING_PARAMS.DEBUG = 0;
        fprintf('DEBUG: Comparison of estimated gradients:\n')
        for k = 1:NUM_OBSERVATIONS
            if M == 1
                GRADESTI = gradest(@(FPARAMS_T) estimator_family(APOSTPDF_SAMPLES(k), LOG_APOSTPDF_VALS(k), vect2struct(FPARAMS_T), TMP_LEARNING_PARAMS), struct2vect(FPARAMS));                
                disp([GRADIENT.XI1(:,k) GRADIENT.XI2(:,k)  GRADIENT.XI3(:,k) GRADIENT.LAMBDA(:,k)  GRADIENT.PHI(:,k) GRADIENT.PSI(:,k); ...
                    GRADIENT_.XI1(:,k)  GRADIENT_.XI2(:,k) GRADIENT_.XI3(:,k) GRADIENT_.LAMBDA(:,k) GRADIENT_.PHI(:,k) GRADIENT_.PSI(:,k);...
                    GRADESTI])
            else
                GRADESTI = gradest(@(FPARAMS_T) select(estimator_family(APOSTPDF_SAMPLES(k), LOG_APOSTPDF_VALS(k), vect2struct(FPARAMS_T), TMP_LEARNING_PARAMS),LEARNING_PARAMS.CHECK_EF_GRADIENT_INDEX), struct2vect(FPARAMS));                
                disp([GRADIENT.XI1(LEARNING_PARAMS.CHECK_EF_GRADIENT_INDEX,k)  GRADIENT.XI2(LEARNING_PARAMS.CHECK_EF_GRADIENT_INDEX,k)  GRADIENT.XI3(LEARNING_PARAMS.CHECK_EF_GRADIENT_INDEX,k)  GRADIENT.LAMBDA(LEARNING_PARAMS.CHECK_EF_GRADIENT_INDEX,k)  GRADIENT.PHI(LEARNING_PARAMS.CHECK_EF_GRADIENT_INDEX,k)  GRADIENT.PSI(LEARNING_PARAMS.CHECK_EF_GRADIENT_INDEX,k); ...
                      GRADIENT_.XI1(LEARNING_PARAMS.CHECK_EF_GRADIENT_INDEX,k) GRADIENT_.XI2(LEARNING_PARAMS.CHECK_EF_GRADIENT_INDEX,k) GRADIENT_.XI3(LEARNING_PARAMS.CHECK_EF_GRADIENT_INDEX,k) GRADIENT_.LAMBDA(LEARNING_PARAMS.CHECK_EF_GRADIENT_INDEX,k) GRADIENT_.PHI(LEARNING_PARAMS.CHECK_EF_GRADIENT_INDEX,k) GRADIENT_.PSI(LEARNING_PARAMS.CHECK_EF_GRADIENT_INDEX,k);...
                    GRADESTI])
            end
        end
        pause
    end
    fprintf('\n')
end

end

%% SUBFUNCTIONS
function [fval,deriv_diag] = f1(Z, XI1, PHI)
if PHI ~= 0.0
    fval = XI1 * Z + bsxfun(@times,PHI,log(abs(Z)));
    if nargout > 1,deriv_diag = XI1 + bsxfun(@rdivide,PHI,Z);end
else
    fval = XI1 * Z;
    if nargout > 1,deriv_diag = XI1;end
end
end



function fval = f2(Z, XI2, XI3, PSI)
    fval = XI2 * Z + XI3 * exp(bsxfun(@times,PSI,Z));
end

function fval = select(x,ind)
fval = x(ind);
end

function [THETA_HAT,GRADIENT] = internal_estimator_family_core(APOSTPDF_SAMPLES,LOG_APOSTPDF_VALS,FPARAMS,M)

NUM_OBSERVATIONS = numel(APOSTPDF_SAMPLES);

THETA_HAT = zeros(M,NUM_OBSERVATIONS);
GRADIENT.XI1 = zeros(M,NUM_OBSERVATIONS);
GRADIENT.XI2 = zeros(M,NUM_OBSERVATIONS);
GRADIENT.LAMBDA = zeros(M,NUM_OBSERVATIONS);
GRADIENT.PHI = zeros(M,NUM_OBSERVATIONS); % store only diagonal of gradient matrix
GRADIENT.PSI = zeros(M,NUM_OBSERVATIONS); % store only diagonal of gradient matrix

for NUM_O = 1:NUM_OBSERVATIONS
    WEIGHTS = LOG_APOSTPDF_VALS{NUM_O}*(FPARAMS.LAMBDA - 1);
    WEIGHTS = WEIGHTS - max(WEIGHTS);
    WEIGHTS = exp(WEIGHTS);
    WEIGHTS = WEIGHTS' / sum(WEIGHTS); % make weights a column vector, which is already normalized

    FVAL_F2 = f2(APOSTPDF_SAMPLES{NUM_O}, FPARAMS.XI2, FPARAMS.XI3, FPARAMS.PSI);
    ARGUMENT_F1 = FVAL_F2 * WEIGHTS;

    [FVAL_F1,DERIV_F1_DIAG] = f1( ARGUMENT_F1, FPARAMS.XI1, FPARAMS.PHI);
    THETA_HAT(:,NUM_O) = FVAL_F1;

    if nargout > 1
        GRADIENT.XI1(:,NUM_O) = ARGUMENT_F1;
        GRADIENT.XI2(:,NUM_O) = bsxfun(@times,DERIV_F1_DIAG, ...
            APOSTPDF_SAMPLES{NUM_O} * WEIGHTS );
        GRADIENT.XI3(:,NUM_O) = bsxfun(@times,DERIV_F1_DIAG, ...
            exp(bsxfun(@times,FPARAMS.PSI,APOSTPDF_SAMPLES{NUM_O})) * WEIGHTS );
        GRADIENT.LAMBDA(:,NUM_O) = bsxfun(@times,DERIV_F1_DIAG,bsxfun(@times,FVAL_F2,LOG_APOSTPDF_VALS{NUM_O}) * WEIGHTS ...
            - ARGUMENT_F1 * (LOG_APOSTPDF_VALS{NUM_O}*WEIGHTS));
        GRADIENT.PHI(:,NUM_O) = log(abs(ARGUMENT_F1));
        GRADIENT.PSI(:,NUM_O) = FPARAMS.XI3*bsxfun(@times,DERIV_F1_DIAG,(exp(bsxfun(@times,FPARAMS.PSI,APOSTPDF_SAMPLES{NUM_O})).*APOSTPDF_SAMPLES{NUM_O}) * WEIGHTS);
    end
end
end

